# -*- coding: utf-8 -*-
"""
w2lapp.modify: modify an entry

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)
"""

import ldap,ldif,ldap.schema,ldaputil.schema, \
       ldapsession,w2lapp.core,w2lapp.cnf,w2lapp.gui,w2lapp.addmodifyform,w2lapp.add

from w2lapp.schema.syntaxes import syntax_registry,LDAPSyntaxValueError

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO

from ldap.schema.models import AttributeType
from ldaputil.modlist2 import modifyModlist

def GetEntryfromInputForm(form,ls,dn,sub_schema):

  entry = ldaputil.schema.Entry(sub_schema,dn.encode(ls.charset),{})
  try:
    in_ldif = form.field['in_ldif'].getLDIFRecords()
  except ValueError,e:
    raise w2lapp.core.ErrorExit(u'LDIF parsing error: %s' % (str(e)))
  else:
    if in_ldif:
      entry.update(in_ldif[0][1])

  # Get all the attribute types
  in_attrtype_list = [
    a.encode('ascii')
    for a in form.getInputValue('in_at',[])
  ]
  # Grab the raw input strings
  in_value_indexes = [
    a for a in form.getInputValue('in_avi',[])
  ]
  # Grab the raw input strings
  in_value_list = [
    a for a in form.getInputValue('in_av',[])
  ]

  if not (len(in_attrtype_list)==len(in_value_list)==len(in_value_indexes)):
    raise w2lapp.core.ErrorExit(u'Different count of attribute types and values input.')

  for i,attr_type in enumerate(in_attrtype_list):
    attr_instance = syntax_registry.attrInstance(None,form,ls,dn,sub_schema,attr_type,None,entry=entry)
    try:
      attr_value = attr_instance.sanitizeInput(in_value_list[i])
    except LDAPSyntaxValueError:
      attr_value = in_value_list[i]
    try:
      entry[attr_type].append(attr_value)
    except KeyError:
      entry[attr_type] = [attr_value]

  iteration_count = 7
  entry_changed = True
  while entry_changed and iteration_count:
    iteration_count -= 1
    entry_changed = False
    # Sanitizing all values after filling the entry
    for attr_type,attr_values in entry.items():
      attr_instance = syntax_registry.attrInstance(None,form,ls,dn,sub_schema,attr_type,None,entry=entry)
      try:
        new_values = attr_instance.transmute(attr_values)
      except (KeyError,IndexError):
        entry_changed = True
        entry[attr_type] = ['']
      else:
        entry_changed = entry_changed or (new_values!=attr_values)
        entry[attr_type] = new_values

  invalid_attrs = []

  # Checking for invalid input done after sanitizing all values so
  # plugin classes can use all entry's attributes for cross-checking input
  for attr_type,attr_values in entry.items():
    attr_instance = syntax_registry.attrInstance(None,form,ls,dn,sub_schema,attr_type,None,entry=entry)
    for attr_value in attr_values:
      if attr_value:
        try:
          attr_instance.validate(attr_value)
        except LDAPSyntaxValueError:
          invalid_attrs.append(attr_type)

  return entry,sorted(set(invalid_attrs)) # GetEntryfromInputForm()


def ModlistLDIF(dn,form,modlist):
  """
  Return a string containing a HTML-formatted LDIF change record
  """
  s = []
  s.append('<code>')
  f = StringIO()
  ldif_writer = ldif.LDIFWriter(f)
  ldif_writer.unparse(dn.encode('utf-8'),modlist)
  s.append(form.utf2display(f.getvalue().decode('utf-8')).replace('\n','<br>'))
  s.append('</code>')
  return ''.join(s) # ModlistTable()


##############################################################################
# Modify existing entry
##############################################################################

def w2l_Modify(sid,outf,command,form,ls,dn):

  sub_schema = ls.retrieveSubSchema(
    dn,
    w2lapp.cnf.GetParam(ls,'_schema',None),
    w2lapp.cnf.GetParam(ls,'supplement_schema',None),
    w2lapp.cnf.GetParam(ls,'schema_strictcheck',True),
  )

  in_assertion = form.getInputValue('in_assertion',[u'(objectClass=*)'])[0]

  input_modrow = form.getInputValue('in_mr',['.'])[0]

  if input_modrow[0]=='-':
    del_row_num = int(input_modrow[1:])
    del form.field['in_at'].value[del_row_num]
    del form.field['in_av'].value[del_row_num]
    # FIX ME! This is definitely not sufficient!
    del form.field['in_avi'].value[del_row_num]
  elif input_modrow[0]=='+':
    insert_row_num = int(input_modrow[1:])
    form.field['in_at'].value.insert(insert_row_num+1,form.field['in_at'].value[insert_row_num])
    form.field['in_av'].value.insert(insert_row_num+1,'')
    # FIX ME! This is definitely not sufficient!
    form.field['in_avi'].value.insert(insert_row_num+1,form.field['in_avi'].value[insert_row_num])

  new_entry,invalid_attrs = w2lapp.modify.GetEntryfromInputForm(form,ls,dn,sub_schema)

  # Check if the user just switched/modified input form
  if 'in_ft' in form.inputFieldNames or \
     'in_oc' in form.inputFieldNames or \
     'in_mr' in form.inputFieldNames or \
     not new_entry or \
     invalid_attrs:

    if invalid_attrs:
      error_msg = 'Syntax check failed for the following attributes: %s' % (
        ', '.join([
          '<a class="CommandLink" href="#in_a_%s">%s</a>' % (v,v)
          for v in map(form.utf2display,map(unicode,invalid_attrs))
        ])
      )
    else:
      error_msg = ''
    w2lapp.addmodifyform.w2l_ModifyForm(
      sid,outf,'modify',form,ls,dn,
      new_entry,
      Msg=error_msg,
    )
    return

  in_oldattrtypes = {}
  for a in form.getInputValue('in_oldattrtypes',[]):
    attr_type = a.encode('ascii')
    in_oldattrtypes[attr_type] = None

  try:
    old_entry,dummy = w2lapp.addmodifyform.ReadOldEntry(ls,dn,sub_schema,in_assertion)
  except ldap.NO_SUCH_OBJECT:
    raise w2lapp.core.ErrorExit(u'Old entry was removed or modified in between! You have to edit it again.')

  # Filter out empty values
  for attr_type,attr_values in new_entry.items():
    new_entry[attr_type] = filter(None,attr_values)

  # Set up a dictionary of all attribute types to be ignored
  ignore_attr_types = ldap.cidict.cidict(
    {}.fromkeys([
      sub_schema.getoid(AttributeType,at_name,raise_keyerror=0)
      for at_name in w2lapp.add.ADD_IGNORE_ATTR_TYPES
    ])
  )

  # Determine whether Relax Rules control is in effect
  relax_rules_enabled = ls.l._get_server_ctrls('**write**').has_key(ldapsession.CONTROL_RELAXRULES)

  if not relax_rules_enabled:
    # Add all attributes which have NO-USER-MODIFICATION set
    ignore_attr_types.update(sub_schema.no_user_mod_attr_oids)
    # Ignore attributes which are assumed to be constant (some operational attributes)
    ignore_attr_types.update(w2lapp.addmodifyform.ConfiguredConstantAttributes(ls))

  # All attributes currently read which were not viewable before
  # must be ignored to avoid problems with different access rights
  # after re-login
  ignore_attr_types.update({}.fromkeys([
    a
    for a in old_entry.keys()
    if not in_oldattrtypes.has_key(a)
  ]))

  # Ignore binary attributes from old entry data in any case
  for attr_type in old_entry.keys():
    syntax_class = syntax_registry.syntaxClass(sub_schema,attr_type)
    if not syntax_class.editable:
      ignore_attr_types[sub_schema.getoid(AttributeType,attr_type,raise_keyerror=0)] = None

  try:
    del ignore_attr_types['2.5.4.0']
  except KeyError:
    pass

  # Create modlist containing deltas
  modlist = modifyModlist(
    sub_schema,
    old_entry,new_entry,
    ignore_attr_types=ignore_attr_types.keys(),
    ignore_oldexistent=0,
  )
  # Binary values are always replaced
  for attr_type in new_entry.keys():
    syntax_class = syntax_registry.syntaxClass(sub_schema,attr_type)
    if (not syntax_class.editable) and \
       new_entry[attr_type] and \
       (not attr_type in old_entry or new_entry[attr_type]!=old_entry[attr_type]):
      modlist.append((ldap.MOD_REPLACE,attr_type,new_entry[attr_type]))

  if modlist:
    # Send modify-list to host
    try:
      ls.modifyEntry(dn,modlist,assertion_filter=in_assertion)
    except ldap.ASSERTION_FAILED:
      raise w2lapp.core.ErrorExit(u'Old entry was removed or modified in between! You have to edit it again.')
    except (
      ldap.CONSTRAINT_VIOLATION,
      ldap.INVALID_DN_SYNTAX,
      ldap.INVALID_SYNTAX,
      ldap.NAMING_VIOLATION,
      ldap.OBJECT_CLASS_VIOLATION,
      ldap.OTHER,
      ldap.TYPE_OR_VALUE_EXISTS,
      ldap.UNDEFINED_TYPE,
      ldap.UNWILLING_TO_PERFORM,
    ),e:
      w2lapp.addmodifyform.w2l_ModifyForm(
        sid,outf,'modify',form,ls,dn,
        new_entry,
        Msg=w2lapp.gui.LDAPError2ErrMsg(e,form,ls.charset),
      )
      return
    else:
      # delete all cache entries referencing to old DN
      UserMsg = '<p class="SuccessMessage">Modified entry %s</p><dt>LDIF change record:</dt>\n<dd>%s</dd>' % (
        w2lapp.gui.DisplayDN(sid,form,ls,dn),
        ModlistLDIF(dn,form,modlist),
      )
  else:
    UserMsg = '<p class="SuccessMessage">No attributes modified of entry %s</p>' % (w2lapp.gui.DisplayDN(sid,form,ls,dn))

  # Output comes here
  w2lapp.gui.SimpleMessage(
    sid,outf,command,form,ls,dn,
    'Modify result',
    UserMsg,
    main_menu_list=w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )

